open Instr
open Typer
open Ast_typer
open Ast

let nyi x = Printf.eprintf "Error: `%s' not yet implemented\n%!" x

let initialise_code () =
  Prim.open_descriptor_in 0;
  Code.add (Setglobal Gentools.stdin);
  Prim.open_descriptor_out 1;
  Code.add (Setglobal Gentools.stdout);
  Prim.open_descriptor_out 2;
  Code.add (Setglobal Gentools.stderr);
;;

let finalise_code () =
  Prim.flush Gentools.stdout;
  Prim.flush Gentools.stderr;
  Code.add Stop;
;;

let rec gen_unary tc arg_nb method_sofs block_sizes expr_sofs t op e =
  gen_expr tc arg_nb method_sofs block_sizes expr_sofs e;
  match (t, op) with
    | `byte , Oneg | `short, Oneg | `char, Oneg | `int, Oneg ->
	Code.add Negint;
    | `float, Oneg ->
	Code.add (Ccall (1, Prim.add "caml_neg_float"));
    | `boolean, Onot ->
	Code.add Boolnot;
    | `byte , Otild | `short, Otild | `char, Otild | `int, Otild ->
	Code.add Push;
	Code.add (Const (-1));
	Code.add Xorint;
    | `byte, Opreinc | `short, Opreinc | `char, Opreinc | `int, Opreinc ->
	begin match fst e with
	  | `byte | `short | `char | `int ->
	      begin match snd e with
		| Ast_typer.Evar (Ast_typer.Alocal (_, bid, vid)) ->
		    let var_ind =
		      Genenv.compute_var_ind block_sizes expr_sofs bid vid
		    in
		      Code.add (Offsetint 1);
		      Code.add (Assign var_ind);
		| _ -> nyi "Opreinc"
	      end;
	  | _ -> nyi "Opreinc"
	end;
    | `byte, Opredec | `short, Opredec | `char, Opredec | `int, Opredec ->
	begin match fst e with
	  | `byte | `short | `char | `int ->
	      begin match snd e with
		| Ast_typer.Evar (Ast_typer.Alocal (_, bid, vid)) ->
		    let var_ind =
		      Genenv.compute_var_ind block_sizes expr_sofs bid vid
		    in
		      Code.add (Offsetint (-1));
		      Code.add (Assign var_ind);
		| _ -> nyi "Opredec"
	      end;
	  | _ -> nyi "Opredec"
	end;
    | `byte, Opostinc | `short, Opostinc | `char, Opostinc | `int, Opostinc ->
	begin match fst e with
	  | `byte | `short | `char | `int ->
	      begin match snd e with
		| Ast_typer.Evar (Ast_typer.Alocal (_, bid, vid)) ->
		    let var_ind =
		      Genenv.compute_var_ind block_sizes expr_sofs bid vid
		    in
		      Code.add (Offsetint 1);
		      Code.add (Assign var_ind);
		      Code.add (Offsetint (-1));
		| _ -> nyi "Opostinc"
	      end;
	  | _ -> nyi "Opostinc"
	end;
    | `byte, Opostdec | `short, Opostdec | `char, Opostdec | `int, Opostdec ->
	begin match fst e with
	  | `byte | `short | `char | `int ->
	      begin match snd e with
		| Ast_typer.Evar (Ast_typer.Alocal (_, bid, vid)) ->
		    let var_ind =
		      Genenv.compute_var_ind block_sizes expr_sofs bid vid
		    in
		      Code.add (Offsetint (-1));
		      Code.add (Assign var_ind);
		      Code.add (Offsetint 1);
		| _ -> nyi "Opostdec"
	      end;
	  | _ -> nyi "Opostdec"
	end;
    | _ ->
	assert false

and gen_binary tc arg_nb method_sofs block_sizes expr_sofs t op e1 e2 =
  begin match op with
    | Oandand | Ooror -> ()
    | _ ->
	gen_expr tc arg_nb method_sofs block_sizes expr_sofs e2;
	Code.add Push;
	gen_expr tc arg_nb (succ method_sofs) block_sizes (succ expr_sofs) e1;
  end;
  match t with
    | `byte | `short | `char | `int | `long ->
	Code.add (
	  match op with
	    | Oplus -> Addint
	    | Ominus -> Subint
	    | Omul -> Mulint
	    | Odiv -> Divint
	    | Omod -> Modint
	    | Oor -> Orint
	    | Oand -> Andint
	    | Oxor -> Xorint
	    | Oinfinf -> Lslint
	    | Osupsup -> Lsrint
	    | Osupsupsup -> Asrint
	    | _ -> assert false
	)
    | `boolean ->
	begin match op with
	  | Oor -> Code.add Orint
	  | Oand -> Code.add Andint
	  | Oxor -> Code.add Xorint
	  | Oandand ->
	      let ptr = Code.create_addr () in
		gen_expr tc arg_nb method_sofs block_sizes expr_sofs e1;
		Code.add (Branchifnot ptr);
		gen_expr tc arg_nb method_sofs block_sizes expr_sofs e2;
		Code.register_addr ptr;
	  | Ooror ->
	      let ptr = Code.create_addr () in
		gen_expr tc arg_nb method_sofs block_sizes expr_sofs e1;
		Code.add (Branchif ptr);
		gen_expr tc arg_nb method_sofs block_sizes expr_sofs e2;
		Code.register_addr ptr;
	  | _ -> assert false
	end
    | `float | `double ->
	Code.add (Ccall (2, Prim.add (
			   match op with
			     | Oplus -> "caml_add_float"
			     | Ominus -> "caml_sub_float"
			     | Omul -> "caml_mul_float"
			     | Odiv -> "caml_div_float"
			     | _ -> assert false
			 )));
    | `string ->
	begin match op with
	  | Oplus ->
	      Gentools.gen_concat_string ();
	  | _ -> assert false
	end
    | `array _ -> assert false
    | `java_object _ -> assert false
    | `void -> assert false
    | `nil -> assert false

and gen_bincomp tc arg_nb method_sofs block_sizes expr_sofs t op e1 e2 =
  let t = fst e1 in
    assert (t = fst e2);
    gen_expr tc arg_nb method_sofs block_sizes expr_sofs e2;
    Code.add Push;
    gen_expr tc arg_nb (succ method_sofs) block_sizes (succ expr_sofs) e1;
    match t with
      | `int | `short | `byte | `char | `long ->
	  begin match op with
	    | Oinf      -> Code.add Ltint;
	    | Oinfequal -> Code.add Leint;
	    | Osup      -> Code.add Gtint;
	    | Osupequal -> Code.add Geint;
	    | Oequal    -> Code.add Eq;
	    | Onotequal -> Code.add Neq;
	  end
      | `float | `double ->
	  Code.add (Ccall (2, Prim.add (
			     match op with
			       | Oinf      -> "caml_lt_float"
			       | Oinfequal -> "caml_le_float"
			       | Osup      -> "caml_gt_float"
			       | Osupequal -> "caml_ge_float"
			       | Oequal    -> "caml_eq_float"
			       | Onotequal -> "caml_neq_float"
			   )));
      | `string | `java_object _ | `array _ | `boolean ->
	  begin match op with
	    | Oequal    -> Code.add Eq;
	    | Onotequal -> Code.add Neq;
	    | _ -> assert false
	  end
      | `void -> assert false
      | `nil  -> assert false
	  
and gen_expr tc arg_nb method_sofs block_sizes expr_sofs e =
  match snd e with
    | Ast_typer.Eunary (op, e0) ->
	gen_unary tc arg_nb method_sofs block_sizes expr_sofs (fst e) op e0;
    | Ast_typer.Ebinary (op, e1, e2) ->
	gen_binary tc arg_nb method_sofs block_sizes expr_sofs (fst e) op e1 e2;
    | Ast_typer.Ebincomp (op, e1, e2) ->
	gen_bincomp tc arg_nb method_sofs block_sizes expr_sofs (fst e) op e1 e2
    | Ast_typer.Eternary (Ast.Oifthenelse, e1, e2, e3) ->
	let ptr0 = Code.create_addr () in
	let ptr1 = Code.create_addr () in
	  gen_expr tc arg_nb method_sofs block_sizes expr_sofs e1;
	  Code.add (Branchifnot ptr0);
	  gen_expr tc arg_nb method_sofs block_sizes expr_sofs e2;
	  Code.add (Branch ptr1);
	  Code.register_addr ptr0;
	  gen_expr tc arg_nb method_sofs block_sizes expr_sofs e3;
	  Code.register_addr ptr1;
    | Ast_typer.Ecast(cast, e0) ->
	gen_expr tc arg_nb method_sofs block_sizes expr_sofs e0;
	Gentools.gen_cast cast;
    | Ast_typer.Ebool b ->
	Code.add (Const (if b then 1 else 0));
    | Ast_typer.Eint (i) ->
	Code.add (Const i);
    | Ast_typer.Efloat (f) ->
	Code.add (Getglobal (Data.create f));
    | Ast_typer.Estring (s) ->
	Code.add (Getglobal (Data.create s));
    | Ast_typer.Enew_object (c, id, args) ->
	let cname = List.hd c in
	let switch_addr = Genenv.get_addr_of_class_name cname in
	let nb_method = Genenv.get_nb_method cname in
	let nb_field = Genenv.compute_nb_field cname in
	let constr_arg_nb = List.length args in
	let end_addr = Code.create_addr () in
	let expr_offset = ref expr_sofs in
	let method_offset = ref method_sofs in
	let gen_arg arg =
	  gen_expr tc arg_nb !method_offset block_sizes !expr_offset arg;
	  Code.add Push;
	  incr expr_offset;
	  incr method_offset;
	in
	  if constr_arg_nb > 2 then (
	    Code.add (Pushretaddr end_addr);
	    expr_offset := !expr_offset + 3;
	    method_offset := !method_offset + 3;
	  );
	  List.iter gen_arg args;
	  Code.add (Const (id + nb_method));
	  Code.add (Push);
	  for i = 2 to nb_field do
	    Code.add (Const 0);
	    Code.add (Push);
	  done;
	  if nb_field > 0 then Code.add (Const 0);
	  Code.add (Closure (nb_field, switch_addr));
	  Code.add (Apply (succ constr_arg_nb));
	  Code.register_addr end_addr;
    | Ast_typer.Enew_array (t, el, a) ->
	nyi "Enew_array"
    | Ast_typer.Efield_access (e0, name, id) ->
	gen_expr tc arg_nb method_sofs block_sizes expr_sofs e0;
	Code.add (Getfield (succ id));
    | Ast_typer.Ecall (name, id, args) ->
	let method_arg_nb = List.length args in
	let end_addr = Code.create_addr () in
	let expr_offset = ref expr_sofs in
	let method_offset = ref method_sofs in
	let gen_arg arg =
	  gen_expr tc arg_nb !method_offset block_sizes !expr_offset arg;
	  Code.add Push;
	  incr expr_offset;
	  incr method_offset;
	in
	  if method_arg_nb > 2 then (
	    Code.add (Pushretaddr end_addr);
	    expr_offset := !expr_offset + 3;
	    method_offset := !method_offset + 3;
	  );
	  List.iter gen_arg args;
	  Code.add (Const id);
	  Code.add (Push);
	  incr expr_offset;
	  incr method_offset;
	  Code.add (Acc !method_offset);
	  Code.add (Apply (succ method_arg_nb));
	  Code.register_addr end_addr;
    | Ast_typer.Earray_access (e1, e2) ->
	nyi "Earray_access"
    | Ast_typer.Etype_cast (t, a, e0) ->
	nyi "Etype_cast"
    | Ast_typer.Eexpr_cast (e1, e2) ->
	nyi "Eexpr_cast"
    | Ast_typer.Einstanceof (e, t) ->
	Code.add (Const 0);
	nyi "Einstanceof"
    | Ast_typer.Eaffect (ta, e0) ->
	begin match ta with
	  | Ast_typer.Alocal (_, bid, vid) ->
	      let var_ind =
		Genenv.compute_var_ind block_sizes expr_sofs bid vid
	      in
		gen_expr tc arg_nb method_sofs block_sizes expr_sofs e0;
		Code.add (Assign var_ind);
	  | Ast_typer.Afield (t, cid, fid) ->
	      let field_ind = Genenv.compute_field_ind tc cid fid in
		gen_expr tc arg_nb method_sofs block_sizes expr_sofs e0;
		Code.add (Push);
		Code.add (Acc (succ method_sofs));
		Code.add (Setfield field_ind);
	  | Ast_typer.Aparam (t, id) ->
	      gen_expr tc arg_nb method_sofs block_sizes expr_sofs e0;
	      Code.add (Assign (method_sofs + 1 + arg_nb - id));
	  | Ast_typer.Acplxname names -> nyi "Eaffect (Acplxname)"
	  | Ast_typer.Acplxfield (e0, name) -> nyi "Eaffect (Acplxfield)"
	  | Ast_typer.Acplxarray (tabe, ind) -> nyi "Eaffect (Acplxarray)"
	end
    | Ast_typer.Ethis -> Code.add (Acc method_sofs)
    | Ast_typer.Evar ta ->
	begin match ta with
	  | Ast_typer.Alocal (_, bid, vid) ->
	      let var_ind =
		Genenv.compute_var_ind block_sizes expr_sofs bid vid
	      in
		Code.add (Acc var_ind);
	  | Ast_typer.Afield (t, cid, fid) ->
	      let field_ind = Genenv.compute_field_ind tc cid fid in
		Code.add (Envacc field_ind);
	  | Ast_typer.Aparam (t, id) ->
	      Code.add (Acc (method_sofs + 1 + arg_nb - id));
	  | Ast_typer.Acplxname names -> nyi "Evar (Acplxname)"
	  | Ast_typer.Acplxfield (e0, name) -> nyi "Evar (Acplxfield)"
	  | Ast_typer.Acplxarray (tabe, ind) -> nyi "Evar (Acplxarray)"
	end
    | Ast_typer.Enull ->
	Code.add (Const 0)
    | Ast_typer.Esuper ->
	nyi "Esuper"
    | Ast_typer.Eexpr_call (obj, meth, id, args) ->
	let method_arg_nb = List.length args in
	let end_addr = Code.create_addr () in
	let expr_offset = ref expr_sofs in
	let method_offset = ref method_sofs in
	let gen_arg arg =
	  gen_expr tc arg_nb !method_offset block_sizes !expr_offset arg;
	  Code.add Push;
	  incr expr_offset;
	  incr method_offset;
	in
	  if method_arg_nb > 2 then (
	    Code.add (Pushretaddr end_addr);
	    expr_offset := !expr_offset + 3;
	    method_offset := !method_offset + 3;
	  );
	  List.iter gen_arg args;
	  Code.add (Const id);
	  Code.add (Push);
	  incr expr_offset;
	  incr method_offset;
	  gen_expr tc arg_nb !method_offset block_sizes !expr_offset obj;
	  Code.add (Apply (succ method_arg_nb));
	  Code.register_addr end_addr;

and gen_instr tc arg_nb continues breaks method_sofs block_sizes =
  let rec gen_loc_instr label instr =
    match instr with
      | Ast_typer.Ilabeled (l, b) -> gen_loc_instr (Some l) b
      | Ast_typer.Iexpr e -> gen_expr tc arg_nb method_sofs block_sizes 0 e
      | Ast_typer.Idecl (t, name, init) ->
	  begin match init with
	    | Ast_typer.Simple_init e ->
		gen_expr tc arg_nb method_sofs block_sizes 0 e
	    | Ast_typer.Array_init el -> nyi "Idecl (Array_init)"
	  end;
	  Code.add Push;
      | Ast_typer.Iblock (decls, body) ->
	  let block_size = ref 0 in
	  let method_offset = ref method_sofs in
	  let gen_decl d =
	    gen_instr tc arg_nb continues breaks !method_offset
	      (!block_size :: block_sizes) None d;
	    incr block_size;
	    incr method_offset;
	  in
	    List.iter gen_decl decls;
	    List.iter
	      (gen_instr tc arg_nb continues breaks !method_offset
		 (!block_size :: block_sizes) None) body;
	    if !block_size <> 0 then Code.add (Pop !block_size);
      | Ast_typer.Iifthenelse (test, th, el) ->
	  let ptr0 = Code.create_addr () in
	  let ptr1 = Code.create_addr () in
	    gen_expr tc arg_nb method_sofs block_sizes 0 test;
	    Code.add (Branchifnot ptr0);
	    gen_loc_instr None th;
	    Code.add (Branch ptr1);
	    Code.register_addr ptr0;
	    gen_loc_instr None el;
	    Code.register_addr ptr1;
      | Ast_typer.Iswitch (expr, cases) -> nyi "Iswitch"
      | Ast_typer.Iwhile (test, body) ->
	  let ptr0 = Code.create_addr () in
	  let ptr1 = Code.create_addr () in
	    Code.register_addr ptr0;
	    gen_expr tc arg_nb method_sofs block_sizes 0 test;
	    Code.add (Branchifnot ptr1);
	    gen_instr tc arg_nb ((label, method_sofs, ptr0) :: continues)
	      ((label, method_sofs, ptr1) :: breaks) method_sofs block_sizes
	      None body;
	    Code.add (Branch ptr0);
	    Code.register_addr ptr1;
      | Ast_typer.Idowhile (body, test) ->
	  let ptr0 = Code.create_addr () in
	  let ptr1 = Code.create_addr () in
	    Code.register_addr ptr0;
	    gen_instr tc arg_nb ((label, method_sofs, ptr0) :: continues)
	      ((label, method_sofs, ptr1) :: breaks) method_sofs block_sizes
	      None body;
	    gen_expr tc arg_nb method_sofs block_sizes 0 test;
	    Code.add (Branchif ptr0);
	    Code.register_addr ptr1;
      | Ast_typer.Ibreak lo ->
	  let (_, mso, ptr) =
	    match lo with
	      | None ->
		  begin match breaks with
		    | [] -> failwith ("break outside any loop or switch")
		    | hd :: _ -> hd
		  end
	      | Some l ->
		  try Genenv.search_label lo breaks
		  with Not_found -> failwith ("invalid break label: " ^ l)
	  in
	    if mso <> method_sofs then Code.add (Pop (method_sofs - mso));
	    Code.add (Branch ptr);
      | Ast_typer.Icontinue lo ->
	  let (_, mso, ptr) =
	    match lo with
	      | None ->
		  begin match continues with
		    | [] -> failwith ("continue outside any loop or switch")
		    | hd :: _ -> hd
		  end
	      | Some l ->
		  try Genenv.search_label lo breaks
		  with Not_found -> failwith ("invalid continue label: " ^ l)
	  in
	    if mso <> method_sofs then Code.add (Pop (method_sofs - mso));
	    Code.add (Branch ptr);
      | Ast_typer.Ireturn eo ->
	  begin match eo with
	    | None -> Code.add (Acc method_sofs);
	    | Some e -> gen_expr tc arg_nb method_sofs block_sizes 0 e
	  end;
	  Code.add (Return (method_sofs + arg_nb + 2));
      | Ast_typer.Ithrow expr -> nyi "Ithrow"
      | Ast_typer.Isynchronized (expr, block) -> nyi "Isynchronized"
      | Ast_typer.Itry (block, catches, finally) -> nyi "Itry"
      | Ast_typer.Inop -> ()
      | Ast_typer.Iprintln e ->
	  gen_expr tc arg_nb method_sofs block_sizes 0 e;
	  Gentools.gen_println ();
	  Prim.flush Gentools.stdout;
  in
    gen_loc_instr

and gen_method_decl typed_class {
  typed_method_modifiers = typed_method_modifiers;
  typed_method_return = typed_method_return;
  typed_method_name = typed_method_name;
  typed_method_parameters = typed_method_parameters;
  typed_method_throws = typed_method_throws;
  typed_method_body = typed_method_body;
} =
  let param_nb = List.length typed_method_parameters in
  let start = Code.next_addr () in
    if param_nb > 0 then Code.add (Grab param_nb);
    begin match typed_method_body with
      | None -> ()
      | Some instr -> gen_instr typed_class param_nb [] [] 0 [] None instr;
    end;
    Code.add (Return (param_nb + 2));
    start

and gen_constructor typed_class {
  typed_constructor_modifiers = typed_constructor_modifiers;
  typed_constructor_name = typed_constructor_name;
  typed_constructor_parameters = typed_constructor_parameters;
  typed_constructor_throws = typed_constructor_throws;
  typed_constructor_call = typed_constructor_call;
  typed_constructor_body = typed_constructor_body;
} =
  (*
  Code.add (Acc 0);
  Code.add (Push);
  *)
  let param_nb = List.length typed_constructor_parameters in
  let start = Code.next_addr () in
    if param_nb > 0 then Code.add (Grab param_nb);
    begin match typed_constructor_call with
      | Ast_typer.Cnone -> ()
      | Ast_typer.Csuper (id, args) -> () (* TODO : super() *)
      | Ast_typer.Cthis (id, args) ->
	  let constr_id =
	    id + typed_class.tc_method_count + (
	      match typed_constructor_call with
		| Ast_typer.Csuper (_, _) -> Array.length typed_class.tc_cons
		| _ -> 0
	    )
	  in
	  let cons_arg_nb = List.length args in
	  let end_addr = Code.create_addr () in
	  let expr_offset = ref 0 in
	  let method_offset = ref 0 in
	  let gen_arg arg =
	    gen_expr typed_class param_nb !method_offset [] !expr_offset arg;
	    Code.add Push;
	    incr expr_offset;
	    incr method_offset;
	  in
	    if cons_arg_nb > 2 then (
	      Code.add (Pushretaddr end_addr);
	      expr_offset := !expr_offset + 3;
	      method_offset := !method_offset + 3;
	    );
	    List.iter gen_arg args;
	    Code.add (Const constr_id);
	    Code.add (Push);
	    incr expr_offset;
	    incr method_offset;
	    Code.add (Acc !method_offset);
	    Code.add (Apply (succ cons_arg_nb));
	    Code.register_addr end_addr;
    end;
    begin match typed_constructor_body with
      | None -> ()
      | Some instr ->
	  gen_instr typed_class param_nb [] [] 0 [] None instr;
    end;
    Code.add (Acc 0);
    Code.add (Return (param_nb + 2));
    start

and gen_class_decl typed_class =
  let init_addr = Code.next_addr () in
    List.iter (gen_instr typed_class (-2) [] [] 0 [] None)
      typed_class.tc_class.typed_class_initializers;
    Code.add (Return 0);
    let this_constr_addrs =
      List.map (gen_constructor typed_class) (Array.to_list typed_class.tc_cons)
    in
    let super_constr_addrs =
      match typed_class.tc_father with
	| None -> []
	| Some father ->
	    let infos =
	      Hashtbl.find Genenv.infos_of_class_name father.tc_name
	    in
	      infos.Genenv.constructors
    in
    let all_constr_addrs = this_constr_addrs @ super_constr_addrs in
    let addrs_nb = typed_class.tc_method_count + List.length all_constr_addrs in
    let addrs = Array.make addrs_nb (ref (-1)) in
    let meth_tbl = Hashtbl.create 10 in
    let rec fill_constr_addrs l i =
      match l with
	| [] -> ()
	| addr :: tl -> addrs.(i) <- ref addr ; fill_constr_addrs tl (succ i)
    in
    let compute_method_addr (si, id, name, decl) =
      if si = 0 then
	let addr = gen_method_decl typed_class decl in
	  addrs.(id) <- ref addr;
	  Hashtbl.add meth_tbl name addr;
      else
	let meths = Genenv.get_method_addrs typed_class si in
	  addrs.(id) <- ref (Hashtbl.find meths name);
    in
      fill_constr_addrs all_constr_addrs typed_class.tc_method_count;
      Array.iter compute_method_addr typed_class.tc_methods;
      let infos = {
	Genenv.init_addr = init_addr;
	Genenv.methods = meth_tbl;
	Genenv.constructors = this_constr_addrs;
	Genenv.class_addr = Code.next_addr ();
      } in
	Hashtbl.add Genenv.infos_of_class_name typed_class.tc_name infos;
	Code.add (Push);
	Code.add (Acc 1);
	Code.add (Switch (addrs_nb, addrs));
	
and gen_interf_decl _ =
  nyi "gen_interf_decl"

and gen_main_start typed_class =
  let main_filter m = m.typed_method_name = "main" in
  let rec search_mains res tc =
    let cd_mains =
      List.map (fun {typed_method_body = body} -> body)
	(List.filter main_filter tc.tc_class.typed_class_methods)
    in
      cd_mains @ List.fold_left search_mains res tc.tc_children
  in
    match search_mains [] typed_class with
      | [] -> failwith "main method not found"
      | [ Some body ] ->
	  let main = Code.create_addr () in
	  let start = Code.create_addr () in
	    Code.add (Branch start);
	    Code.register_addr main;
	    gen_instr typed_class (-1) [] [] 0 [] None body;
	    Code.add (Return 1);
	    Code.register_addr start;
	    Code.add (Closure (0, main));
	    Code.add (Push);
	    Code.add (Atom 0);
	    Code.add (Push);
	    Code.add (Acc 1);
	    Code.add (Apply 1);
      | _ -> failwith "too main method found"
	  
and gen_initializer_calls () =
  let gen_init_call _ infos =
    let ptr = Code.create_addr () in
      Code.add (Const 0);
      Code.add (Pushretaddr ptr);
      Code.add (Branch (ref infos.Genenv.init_addr));
      Code.register_addr ptr;
  in
    Hashtbl.iter gen_init_call Genenv.infos_of_class_name

      
and gen_compil_unit {
  typed_package = typed_package;
  typed_imports = typed_imports;
  typed_class = typed_class;
  typed_interf_decls = typed_interf_decls;
} =
  let start = Code.create_addr () in
    Genenv.store_class_decls typed_class;
    Code.add (Branch start);
    Genenv.iter_tc gen_class_decl typed_class;
    Code.register_addr start;
    gen_initializer_calls ();
    gen_main_start typed_class;
    Genenv.fill_class_addrs ();
;;
